use crate::config::node::NodeType;
use log::{debug, info, trace};
use std::{env, io};
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use connection::Client;
use domo_proto::commands::node_management::NodeManagement;
use domo_proto::packet::ToPacket;
use domo_proto::packet::identifier::Identifier;
use std::cell::RefCell;

pub mod config;
pub mod connection;

thread_local! {
    static CONFIG: RefCell<config::Config> = RefCell::new(config::Config::from_path(
        env::var("DOMO_CONFIG_PATH")
            .unwrap_or("config.toml".into())
    ).unwrap_or({ info!("Using default config..."); config::Config::default() }))
}

fn main() {
    setup_logging().expect("could not setup logging");

    debug!("Building tokio runtime...");

    let runtime = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()
        .unwrap();

    debug!("Built tokio runtime with all features...");

    info!(
        "Staring as {} node!",
        CONFIG.with_borrow(|c| c.node.node_type.clone())
    );

    match CONFIG.with_borrow(|c| c.node.node_type.clone()) {
        NodeType::Master => runtime.block_on(async {}),
        NodeType::Relay { master_address } => {
            let client = Client::new(master_address.parse().unwrap());
            runtime.block_on(async {
                let mut conn = client.client_connection().await.unwrap();
                let device_id = match CONFIG.with_borrow(|c| c.node.device_id.clone()) {
                    Some(s) => {
                        let device_id = Identifier::from(domo_proto::prelude::as_u32_be(hex::decode(s)
                            .expect("Could not decode hex device_id").as_slice()));
                        let p = conn.send(NodeManagement::RegisterNode {
                            device_id
                        }.to_v1_packet(
                            Identifier::from(0),
                            Identifier::from(0),
                            Identifier::random(),
                            Identifier::from(0)
                        )).await.unwrap();
                        if p.command() == 0x0E {
                            panic!("device id is taken")
                        }
                        trace!("server assigned us: {}", device_id);
                        device_id
                    },
                    None => {
                        trace!("requesting device_id...");
                        let p = conn.send(NodeManagement::RegisterNode {
                            device_id: Identifier::from(0)
                        }.to_v1_packet(
                            Identifier::from(0),
                            Identifier::from(0),
                            Identifier::random(),
                            Identifier::from(0)
                        )).await;
                        Identifier::from(p.unwrap().data().get_data())
                    }
                };
                CONFIG.with_borrow_mut(move |c| c.node.device_id = Some(hex::encode(device_id.bytes)));
            })
        }
        NodeType::Subnet { master_address } => {
            runtime.block_on(async {})
        }
    }
}

fn setup_logging() -> Result<(), fern::InitError> {
    fern::Dispatch::new()
        .format(|out, message, record| match record.metadata().target() {
            _ => out.finish(format_args!(
                "[{} - {}] <{}> {}: {}",
                humantime::format_rfc3339(std::time::SystemTime::now()),
                record.level(),
                record.target(),
                record.file().unwrap_or("??"),
                message
            )),
        })
        .chain({
            fern::Dispatch::new()
                .level(CONFIG.with_borrow(|c| c.log.stdout_level.clone()).into())
                .chain(io::stdout())
        })
        .chain({
            fern::Dispatch::new()
                .level(CONFIG.with_borrow(|c| c.log.stdout_level.clone()).into())
                .chain(fern::log_file("./node.log")?)
        })
        .apply()?;

    Ok(())
}
